/*****************************************************************************
 *
 * Written 2008 by Matthias Kirchner, matze@renhcrik.de
 * with help from Thorsten Erdmann
 * using some routines from W. Reissnegger and B. Walle
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with puliwera; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************************/

#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <glib-2.0/glib.h>
#include <limits.h>
#include <assert.h>

#include "puliwera.h"
#include "display.h"
#include "keyfile.h"
#include "player.h"
#include "info.h"

static char s_config_file[512] = DEFAULT_CONFIG_FILE;
static gsize s_channel_number = 0;
char server_url[40]="";
FILE *mp_pipe;

struct channel {
    char            *name;
    char            *url;
    bool            playlist;
};

static struct channel *s_channels;

static void exit_program(int val)
{
	display_close();
	fprintf(mp_pipe, "quit\n");
	fflush(mp_pipe);
	fclose(mp_pipe);
	exit(val);
}

char player_init(void)
{
	FILE *memoryfile;
	char remembered_url[80];
	int ret;

	system("mplayer -slave -quiet -idle -input file=/tmp/puliwera &");
	mp_pipe=fopen(DEFAULT_PIPE_FILE,"w+");
	if (!mp_pipe) 
	{
		perror("Failed to open pipe");
		return 1;
	}
	else 
	{
		// recall last used station
		memoryfile=fopen(DEFAULT_MEMORY_FILE,"r");
		if (memoryfile) 
		{
			ret = fscanf(memoryfile,"%s",remembered_url);
			if (debug) printf("%s\n",remembered_url);
			fclose(memoryfile);
			fprintf(mp_pipe, "loadfile %s\n", remembered_url);
			fflush(mp_pipe);
		}
		return 0;
	}
}

// by W. Reissnegger
typedef struct menu_d
{
	screen_t parent;
	menu_display_context_t ctx;

#define m_ctx               ((menu_t *) screen)->ctx
#define m_start             ((menu_t *) screen)->ctx.start
#define m_selection         ((menu_t *) screen)->ctx.selection
#define m_view_fullpath     ((menu_t *) screen)->ctx.view_fullpath
#define m_left_margin       ((menu_t *) screen)->ctx.left_margin
}
menu_t;

// by W. Reissnegger
file_entry_t *file_entry_new(const char *fullpath)
{
	file_entry_t *entry;
	entry = (file_entry_t *) malloc(sizeof(file_entry_t));
	assert(NULL != entry);
	entry->fullpath[0] = '\0';
	entry->name[0] = '\0';
	entry->type = 1;
	entry->marked = 0;
	return entry;
}

// by W. Reissnegger
file_entry_t *menu_add(screen_t * screen, const char *name, void (*h) (void *))
{
	file_entry_t *entry;
	entry = file_entry_new(NULL);
	assert(NULL != entry);
	save_strcpy(entry->name, name);
	entry->handler = h;
	m_start = list_add(m_start, entry);
	return entry;
}

// by W. Reissnegger
static void menu_delete(screen_t * screen)
{
	assert(screen);
	list_free(m_start);
	free(screen);
}

// by W. Reissnegger, but modified
static void menu_select_prev(screen_t * screen, int num)
{
	while (num--)
	{
		if (m_selection > 0)
		{
			m_selection--;
		}
	}
}
// by W. Reissnegger, but modified
static void menu_select_next(screen_t * screen, int num)
{
	while (num--)
	{
		if (m_selection < list_len(m_start) - 1)
		{
			m_selection++;
		}
	}
}

// by W. Reissnegger, but modified
static int menu_handle_key(screen_t * screen, int key)
{
	assert(screen);
	switch (key)
	{
	case KEY_DOWN:
		menu_select_next(screen, 1);
		break;
	case KEY_UP:
		menu_select_prev(screen, 1);
		break;
	case KEY_SELECT:
		{
			file_entry_t *entry;
			entry = list_data(list_nth(m_start, m_selection));
			if (entry && entry->handler)
			{
				(entry->handler) (screen);
			}
		}
		break;
	}
	return KEY_HANDLED;
}

// by W. Reissnegger
screen_t *menu_new()
{
	screen_t *screen;

	screen = (screen_t *) malloc(sizeof(menu_t));
	assert(NULL != screen);

	bzero(screen, sizeof(menu_t));
	screen->type		= TYPE_MENU;
	screen->update		= NULL;	// not needed for menu
	screen->delete		= menu_delete;
	screen->handle_key	= menu_handle_key;
	screen->ctx		= &m_ctx;

	return screen;
}

void helpmessage(void)
{
	printf("USAGE: puliwera [-h] [-d <0|1>] [-c configfile]\n");
	printf("	-c specify config file other than /usr/etc/puliwera.conf\n");
	printf("	-d <0|1> with (1) or without (0) debug messages\n");
	printf("	-h display this little help message\n");
}

char strip_url(char *longurl)
{
	int length=0;
	char *dummy;
	char urlbackup[128];

	strcpy(server_url, "                                       ");
	strcpy(urlbackup, longurl);
	dummy = strtok(urlbackup,"/");
	length += strlen(dummy)+1;
	dummy = strtok(NULL,"/");
	length += strlen(dummy)+1;
	strncpy(server_url, longurl, length);
	if (debug) printf("Server: %s\n",server_url);
	return 0;
}

void mplayercall(void *data)
{
	char newinfo[80];
	screen_t *screen = (screen_t *)data;
	screen_t *info;
	FILE *memoryfile;

	if (debug) 
	{
		printf("Name: %i - %s\n", m_selection, s_channels[m_selection].name);
		printf("URL : %i - %s\n", m_selection, s_channels[m_selection].url);
		fflush(stdout);
	}

	// remember selected station for further use
	memoryfile=fopen(DEFAULT_MEMORY_FILE,"w");
	if (memoryfile) 
	{
	fprintf(memoryfile, s_channels[m_selection].url);
	fclose(memoryfile);
	}
	
	strip_url(s_channels[m_selection].url);

	sprintf(newinfo, "Contacting the \"%s\" server:", s_channels[m_selection].name);
	info = info_new(newinfo, 5000);
	info_set_text(info, 1, server_url);
	if (debug) printf("Server-URL: %s\n", server_url);
	display_push(info);
	display_show();

	fprintf(mp_pipe, "loadfile %s\n", s_channels[m_selection].url);
	fflush(mp_pipe);
}

void dummy(void *data)
{}

//partially from B. Walle
void menu_from_channels()
{
	char *string;
	char **channels;
	unsigned int i,j;

	string = key_file_get_string_default("streams", "name", "Internet Radio");
	if (debug) printf("key_file_got_string: %s\n",string);
	channels = key_file_get_keys("streams", &s_channel_number);
	if (debug) printf("key_file_got_channels:\n");

	s_channels = g_new(struct channel, s_channel_number);
	if (!s_channels) {
		printf("Loading key file failed\n");
//		return false;
	}

	screen_t *main_menu = menu_new();

	for (i = 0, j = 0; channels[i]; i++, j++) {
		char *value;
	        if (strcmp(channels[i], "name") == 0) {
			s_channel_number--;
			j--;
			continue;
			}

		s_channels[j].name = g_strdup(channels[i]);
		value = key_file_get_string("streams", s_channels[j].name);
		if (!value) {
			printf("Value %s missing\n",value);
			continue;
			}

		s_channels[j].playlist = g_str_has_prefix(value, "playlist");
		if (s_channels[j].playlist) {
			s_channels[j].url = g_strdup(value + strlen("playlist"));
			g_free(value);
			} 
		else
		s_channels[j].url = value;

		g_strchug(s_channels[j].url);

	if (debug) {
		printf("Name: %s\n",s_channels[j].name);
		printf("URL : %s\n",s_channels[j].url);
		}

	menu_add(main_menu, s_channels[j].name, mplayercall);
	}

	display_push(main_menu);
    g_strfreev(channels);
}

#define MAX_FD(_a,_b) ((_a)>(_b)?(_a):(_b))

static void mainloop()
{
	fd_set fds;
	if (debug) printf("main loop\n");
	while(1)
	{
		struct timeval timeout = { 0, 300000 };
		int fd_player = 5;
		int fd_highest_disp;
		int fd_last;

		FD_ZERO(&fds);
		fd_highest_disp = display_fds(&fds);	// Fetch display fds
		fd_last = MAX_FD(fd_highest_disp, fd_player) + 1;

		if (0 < select(fd_last, &fds, NULL, NULL, &timeout))
		{
			display_handle_event(&fds);
		}
	display_show();
	}
}

int main(int argc, char **argv)
{
	int cs, err;

	while (( cs = getopt( argc, argv, "c:d:h" )) > 0)
	{
		switch (cs)
		{
		case 'c':
			strncpy(s_config_file, optarg, 512);
			break;
		case 'd':
			debug=atoi(optarg);
			break;
		case 'h':
		default:
			helpmessage();
			exit(0);
		}
	}

	if (debug) {
	printf("debug: %i\n",debug);
	printf("configfile: %s\n",s_config_file);
	}

	// signal handlers
	signal(SIGINT, exit_program);
	signal(SIGTERM, exit_program);
	signal(SIGHUP, exit_program);
	signal(SIGKILL, exit_program);

	if (0 != init_display()) exit_program(-1);
	if (0 != player_init()) exit_program(-1);

	// by B. Walle
	// read configuration file
	err = key_file_load_from_file(s_config_file);
	if (!err) {
		printf("Loading key file failed");
		return 1;
	}

	menu_from_channels();
	display_show();

	mainloop();

	if (debug) sleep(10);

	exit_program(0);
	return 0;
}
